-- Scenario (this control.lua) globals are in the environment for `/c`
-- commands. map-specific globals are persisted and reloaded, their
-- definitions are kept separate to ease automation (find `npg`)

red=defines.wire_type.red
green=defines.wire_type.green
mfd='marked_for_deconstruction'
local unknown = {}

-- syntax sugar & other dubious sweeteners
local function selfmeta(t) return setmetatable(t,t) end
local min,max,abs = math.min, math.max, math.abs
local floor,ceil = math.floor, math.ceil
local clamp_range=function(n,v,x) return v<n and n or x<v and x or v end
ename={} for k,v in next,defines.events do ename[v]=k end
function default_trace(ev) print(ename[ev.name],serpent.line(ev)) end

psd={comment=false,nocode=true}
function psl(p,k) return print(serpent.line(p,k or psd)) end
function psb(p,k) return print(serpent.block(p,k or psd)) end
function gpsl(p,k) return game.print(serpent.line(p,k or psd)) end

-- handy table meta, htm(t) allows t:insert(1) and for k,v in t
htm = setmetatable(
    {__index=table, __call=function(s,_,k) return next(s,k) end},
    {__call=function(s,t) return setmetatable(t,s) end})

notrace={}

local function C(v) return function() return v end end
allfalse=setmetatable({},{__index=C(false)})
allzero=setmetatable({},{__index=C(0)})

bootstraps={init=1,load=1,configuration_changed=1,nth_tick=1}
function on (s,...)
    local on_s ='on_'..s
    local e = defines.events[on_s]
    if e then
	notrace[e]=true
	return script.on_event(e,...)
    elseif bootstraps[s] then
	return script[on_s](...)
    else -- a custom input
	return script.on_event(s,...)
	end
    end

-- /c locos=fef'locomotive'
-- /c for _,l in next,locos do l.insert'rocket-fuel' end
function fef(t,e)
    if not t then t={} end
    if type(t)=='string' then t={name=t} end
    if type(t[1])=='string' then t.name=t[1] t[1]=t[#t] t[#t]=nil end
    if not t.name and not t.type and e then t.name=e.name end
    if not t.hop then
	return (e and e.surface or game.surfaces[1])
		.find_entities_filtered(t)
	end
    end

default_from_template={
    all={position=1,direction=1,force=1,surface=1}
    ,['underground-belt']={belt_to_ground_type='type'}
}
--
function ce(t,ev)
    local pidx
    if type(t)=='string' then t={name=t} end
    if type(t[1])=='string' then t.name=t[1] t[1]=t[#t] t[#t]=nil  end
    if type(t[1])=='table' then t.position=t[1] t[1]=t[#t] t[#t]=nil  end
    if ev and type(ev)=='table' then
	if type(ev.__self)=='userdata' then
	    -- load up defaults from entity template
	    if not t.position then
		t.fast_replace=true
		t.spill=false
		end
	    for k,_ in next,default_from_template.all do
		if not t[k] then t[k]=ev[k] end
		end
	    for k,v in next,(default_from_template[ev.prototype.type] or {}) do
		if not t[k] then t[v]=ev[k] end
		end
	else
	    -- get player index (for force and surface) from event
	    pidx = ev.player_index
	    end
	end
    local p=pidx and game.players[pidx] or game.player
    t.force = t.force or p and p.force or game.forces.player
    return (p and p.surface or game.surfaces[1]).create_entity(t)
    end

function iod(text) -- "instead of deconstructing"
    return load([[return
	function(ev) local e,p = ev.entity, game.players[ev.player_index]
	    local b1,b2 = e.get_inventory(1),e.get_inventory(2)
	    e.cancel_deconstruction(p.force)
	    ]] .. text .. ' end'
	)()
    end
--     /c on(mfd,iod([[ e.insert{name='iron-plate',count=1e9} ]]))
--     /c on(mfd,iod([[ return b2 and b2.clear() ]])) -- clears assembler input
--[[ runtime setting too distracting outside the scenario, made startup
on('runtime_mod_setting_changed',function(ev)
    if ev.setting=='benchmark-scenario-keep-resources' then
	keep_resources = settings.global['benchmark-scenario-keep-resources'].value
	-- game.print(serpent.line(settings.global['benchmark-scenario-keep-resources']))
	end
    end)
--]]

function instabuild(ev)
    if ev.created_entity.type=='entity-ghost' then
	inserts=ev.created_entity.item_requests
	if inserts then ev.created_entity.item_requests=nil end
	_,e=ev.created_entity.revive()
	for k,v in next,inserts or empty do
	    e.insert{name=k,count=v}
	    end
	end
    end
--
local function instakill_all_transport_lines(e)
    local l
    for x=1,8 do
	l=e.get_transport_line(x)
	if l then l.clear() end
	end
    end
function instakill(ev)
    local e = ev.entity
    if e.prototype.belt_speed then 
	if e.type=='transport-belt' then
	    e.get_transport_line(1).clear()
	    e.get_transport_line(2).clear()
	else
	    pcall(instakill_all_transport_lines,e)
	    end
	end
    ev.entity.destroy()
    end

-- non-persisted globals table consulted by on_init. All lua globals
-- not added to npg before on_init gets around to it will be persisted
-- in the map's `global` by on_init.
-- All globals above here are not map-specific and don't need sync,
-- they're reference material.
local npg = {game=true} for k in next,_G do npg[k]=true end
-- NOTE that lua tables are shallow-copied so saving persists changes
-- automatically, but changes to non-table lua globals are not
-- automatically updated in the map globals -- differences between
-- _G.somestring and global.somestring may not be intentional.

local first_chunk= 'Precharting and paving of %d×%d chunks in progress'
	    .. ' and will interfere with benchmark results for'
	    .. ' about %s minute%s.'
charting = {
    to_range=32
    , chunk_progress = {}
    , chunksrem = 0
    , caught_up_message='background repaving complete!'..
	' Save now and chunk generation won\'t interfere with benchmarking'
}


keep_resources = 
    (settings.startup['benchmark-scenario-keep-resources'] or {}).value
    or false

-- okay.  on_load runs before on_configuration_changed, so on_load has
-- no way of telling what interfaces are present, checking for those in
-- on_load causes the engine to refuse to load the map.
--
-- In multiplayer, the server's on_configuration_changed will run first
-- thing after on_load and can update the map globals properly before
-- the clients fetch them, but that doesn't help the server's on_load
-- set up conditional hooks needed before clients join if the
-- configuration hasn't changed.
--
-- As of 0.16.26 it's possible to check whether you're running
-- multiplayer, and avoid any desync from setting an on_tick handler
-- nobody else is running.
--
-- imo on_configuration_changed "should" be invoked first, in (an
-- expensively set-up facsimile of) the same environment as migration
-- scripts, but going on the response that suggestion got it's too late
-- to change now.
--
-- . . . oh, f. f. s.  on_load can't even check the multiplayer status,
-- so now I have to slather on _more_ dance steps.
--
local function whether(k) return k and '' or 'not ' end
use_bp_insta = unknown
use_bp_cheat = unknown
is_multiplayer = unknown
local function check_configuration(ccdata)
    global.is_multiplayer = game.is_multiplayer()
    is_multiplayer = global.is_multiplayer
    do local _=(remote.interfaces.BenchmarkProduction or {})
	global.use_bp_insta = _.insta or false
	global.use_bp_cheat = _.cheat or false
	use_bp_insta = global.use_bp_insta
	use_bp_cheat = global.use_bp_cheat
	end
    end
on('configuration_changed',check_configuration)
--
local function realconditionalhooks(ev)
    default_trace(ev)
    on(ev.name,nil)
    check_configuration()
    if use_bp_insta then
	for x=1,#game.players do
	    remote.call('BenchmarkProduction','insta'
		    ,{player_index=x})
	    end
    else
	on('marked_for_deconstruction',instakill)
	on('built_entity',instabuild)
	end
    end
local function conditionalhooks()
    if global.is_multiplayer then
	on('player_joined_game',realconditionalhooks)
    else
	on('tick',realconditionalhooks)
	end
    end
on('load',function() 
    for k,v in next,global do _G[k]=v end
    conditionalhooks()
    end)
on('init',function()
    keep_resources = 
	(settings.startup['benchmark-scenario-keep-resources'] or {}).value
	or false
    -- load all persistent globals into the map-saved `globals`
    for k,v in next,_G do if not npg[k] then global[k]=v end end
    conditionalhooks()
    game.surfaces[1].always_day=true
    do local _=game.forces.player
	_.research_all_technologies()
	_.enable_all_recipes()
	_.character_running_speed_modifier = 1.25
	_.character_reach_distance_bonus = 64
	_.character_build_distance_bonus = 64
	_.character_inventory_slots_bonus = 40
	end
    if not keep_resources then
	do local _=game.map_settings
	    _.pollution.enabled = false
	    _.enemy_evolution.enabled = false
	    _.enemy_expansion.enabled = false
	    end
	do local _=game.default_map_gen_settings
	    _.peaceful_mode=true
	    _.water='none'
	    for _,v in next,_.autoplace_controls do
		v.size='none'
		end
	    end
	end
    init_chunks_to(charting.to_range)
    end)

on('player_created',function(ev)
    default_trace(ev)
    local p=game.players[ev.player_index]
    if use_bp_cheat then
	remote.call('BenchmarkProduction','cheat',ev)
    else
	p.cheat_mode = true
	local count = false
	for _,v in next,{1,'power-armor-mk2' ,'steel-axe'
	    ,1000,'rail','train-stop','rail-signal'
	    ,10,'locomotive',30,'cargo-wagon'
	    ,'small-electric-pole','medium-electric-pole','big-electric-pole'
	    ,'substation',1,'electric-energy-interface'
	    ,'arithmetic-combinator','decider-combinator','constant-combinator'
	    ,'red-wire','green-wire'
	    ,'construction-robot'
	    ,'stack-inserter','stack-filter-inserter'
	    ,'express-transport-belt','express-underground-belt','express-splitter'
	    ,'steel-chest'
	    ,'logistic-chest-requester','logistic-chest-active-provider'
	    ,'logistic-chest-passive-provider', 'logistic-chest-storage'
	    ,1,'submachine-gun','uranium-rounds-magazine'
	    ,'rail-chain-signal'
	} do
	    if type(v) == 'number' then
		count = v
	    elseif not game.item_prototypes[v] then
		p.print(string.format('no %s prototype?',v))
		count=false
	    else
		p.insert{name=v, count=count or game.item_prototypes[v].stack_size}
		count=false
		end
	    end

	local grid=p.get_inventory(defines.inventory.player_armor)[1].grid
	for _,v in next,{3,'fusion-reactor-equipment'
		,4,'battery-mk2-equipment'
		,2,'personal-roboport-mk2-equipment'
		,2,'battery-mk2-equipment'
		,2,'personal-roboport-mk2-equipment'
		,12,'battery-mk2-equipment'} do
	    if type(v) == 'number' then
		count = v
	    elseif not game.item_prototypes[v] then
		p.print(string.format('no %s prototype?',v))
		count=false
	    else
		if not count then
		    count=game.item_prototypes[v].stack_size
		    end
		for _=1,count do
		    local e=grid.put{name=v}
		    if e then
			count = count-1
			if v:sub(1,7)=='battery' then
			    e.energy = e.max_energy
			    end
			end
		    end
		if count>0 then p.insert{name=v,count=count} end
		count=false
		end
	    end
	end
    end)

local pkey = function(x,y) return x/2^21+y/2^42 end
local markers={} for x = -301,299,300 do for y=-301,299,300 do
    markers[pkey(x,y)] = 'refined-hazard-concrete-left'
    markers[pkey(x+1,y)] = 'refined-hazard-concrete-right'
    markers[pkey(x,y+1)] = 'refined-hazard-concrete-right'
    markers[pkey(x+1,y+1)] = 'refined-hazard-concrete-left'
    end end
--
local tiles={} for n=1,33*33 do tiles[n]={name='',position={x=0,y=0}} end
function init_chunks_to(N)
    local Nx = 2*N + 1
    local inabout = floor((Nx*Nx/60+15)/60)
    first_chunk_message=first_chunk:format(Nx, Nx
	    ,inabout==1 and 'a' or tostring(inabout)
	    ,inabout==1 and '' or 's')
    do local _=charting
	_.chunksrem = 1--see on_chunk_generated tracking
	for range=1,N do
	    _.chunk_progress[range]=8*range
	    _.chunksrem=_.chunksrem+8*range
	    end
	end
    game.forces.player.chart(game.surfaces[1],{{-32*N,-32*N},{32*N,32*N}})
    end
npg.init_chunks_to = true
--
on('chunk_generated', function(ev)
    if ev.tick==0 then
	game.print(first_chunk_message)
	end
    local range=0
    do local _=ev.area.left_top
	range=max(abs(_.x)/32,abs(_.y)/32)
	end
    if range>0 then
	if range>#charting.chunk_progress then
	    game.forces.player.chart(ev.surface,{{range*-32,range*-32},{range*32,range*32}})
	    game.print(string.format('charting/repaving to range %d',range))
	    end
	if not charting.chunk_progress[range] then
	    charting.chunksrem=charting.chunksrem+8*range
	    --ev.surface.request_to_generate_chunks({0,0},range)
	    end
	charting.chunk_progress[range]=(charting.chunk_progress[range] or 8*range)-1
	end
    for _, entity in next,ev.surface.find_entities(ev.area) or empty do
	if entity.valid and not ( entity.type == 'player'
	   or keep_resources and entity.type == 'resource' ) then
	    entity.destroy()
	    end
        end
    local x,y = ev.area.left_top.x, ev.area.left_top.y
    local n = 1
    for x = x-1, x + 31 do
        for y = y-1, y + 31 do
            tiles[n].name = markers[pkey(x,y)] or 'lab-dark-2'
	    tiles[n].position.x = x
	    tiles[n].position.y = y
	    n = n+1
            end
        end
    ev.surface.set_tiles(tiles,false)
    charting.chunksrem = charting.chunksrem - 1
    if charting.chunksrem == 0 then
	game.forces.player.chart_all()
	game.print(charting.caught_up_message)
	charting.caught_up_message=''
    elseif charting.chunksrem < 0 then
	print('?? chunk',serpent.line(ev.area),'generated')
	end
    end)
on('research_started',function(ev)
--              one-tick delay to keep the auto-research mod happy
--		local t=ev.tick
--		on('tick',function(...) if game.tick>t then
--		    on('tick',nil)
	ev.research.force.research_progress=1
--		    end end)
    end)
-- vim: set sts=4 sw=4 tw=71 lbr fo=tcroq com=\:-- :
